#
# $Id: info.tcl,v 1.18 2000/01/27 16:22:46 elias Exp $
#
# A limited implementation of the catinfo interface.
#
# - Does not handle quoted values.
# - Can only change existing values
# - Can change existing commented-out values, as long as
#   the comment is "simple enough" (a leading # is fine)
# - Will "delete" values by commenting them out
# - Only operates on files 
# - Does not do automagic loading
# - Does very little error-checking.  Almost anything is
#   accepted as a "pathname".  And it is possible to set
#   nodes to values...
#
package require registry 1.0
 

namespace eval ::CatInfo {; #

catch {namespace import CatDbg::*}

namespace export Init Load Flush Set Get PathGet
namespace export Reload LoadFrom List InfoName DeleteNode
namespace export LoadRegFromFile
 

# The global data array. Contains the following kind of elements:
#
#  data,<path> = value
#  del,<path>  = 0/1	exists => modified, 0/1 => deleted / set
#  file,<path> = file	file associated with a node
#  name,<path> = <Path> <path> is lower-case, <Path> is from the file
#
variable info

variable ROOT
variable REGROOT
 

# Init
# -- Always call this one first
#
proc Init {rootdir} {
    variable info
    variable ROOT $rootdir
    
    variable REGROOT {HKEY_LOCAL_MACHINE\SOFTWARE\Security Dynamics\Agents}
    

    D 1 "CatInfo::Init $rootdir"
    catch {unset info}
    set info(dirty) 0

 
    catch {registry keys $REGROOT global} regcode
    if ![string match global $regcode] {
	if [file exists $rootdir/global.cfg] {
	    LoadRegFromFile \\global $rootdir/global.cfg
	} else {
	    return 0
	}
    }

    registry set $REGROOT\\services
    foreach file [glob -nocomplain [file join $rootdir services *.cfg]] {
	regexp {^.*/(.+)\.cfg$} $file {} Key
	regsub -all {\.} $Key {_} Key
	catch {registry keys $REGROOT\\services $Key} regcode
	if ![string match $Key $regcode] {
	    LoadRegFromFile \\services\\$Key $file
	}
    }

    registry set $REGROOT\\tcp

    LoadFrom .global $REGROOT\\global
    foreach file [registry keys $REGROOT\\services] {
	LoadFrom .services.[InfoName $file] $REGROOT\\services\\$file
    }
    foreach file [registry keys $REGROOT\\tcp] {
	LoadFrom .tcp.[InfoName $file] $REGROOT\\tcp\\$file
    }
 
    Reload
}

# LoadFrom
# -- Register loading from some place with Reload
#
proc LoadFrom {Node file} {
    variable info

    set node [string tolower $Node]
    D 1 "CatInfo::LoadFrom $Node ($node) $file"
    set info(file,$node) $file
}

# Reload
# -- Re-read all config files that have been loaded so far
#
proc Reload {} {
    variable ROOT
    variable info

    D 1 "CatInfo::Reload"
    foreach x [array names info file,*] {
	regexp {file,(.*)} $x {} path
	truncate $path
	Load $path $info($x)
    }
}

# Load node file
# -- load a file, and hang its data off a certain node
#
proc Load {Node file} {
    variable info

    set keys [registry keys $file]
    set values [registry values $file]

    set node [string tolower $Node]
    D 1 "CatInfo::Load $Node $file"
    set info(file,[filetonode $node]) $file

    set section .
    set Section .
    foreach Par $values {
	regsub -all {\.} $Par _ Param
	set param [string tolower $Param]
	set info(data,$node$section$param) [registry get $file $Par]
	set info(name,$node$section$param) $Node$Section$Param
	catch { unset info(del,$node$section$param) }
    }

    foreach k $keys {
	Load $Node.$k $file\\$k
    }
}

# LoadRegFromFile node filename
# -- Load registry from file 
#
proc LoadRegFromFile {Node file} {
    variable REGROOT

    D 1 "CatInfo::LoadRegFromFile $Node $file"

    set f [open $file]
    set data [read $f]
    close $f

    if [string match {\\} $Node] {
	set Node {}
    } else {
	if ![string match {\\*} $Node ] {
	    set Key \\$Node
	    set Node $Key
	}
    }

    set Key $REGROOT$Node 
    foreach line [split $data "\n"] {
	set line [string trimright $line]
	if {![string length $line] || [regexp {\s*#.*} $line]} continue
	if [regexp {^(\S[^:]+):$} $line {} name] {
	    regsub -all {\.} $name {\\} name
	    if [string match {\\?*} $name] {
		set Key $REGROOT$name
	    } else {
		set Key $REGROOT$Node\\$name
	    }
	    continue
	}
	set line [string trimleft $line]
	if ![regexp {^(\S+)\s*=\s*(.*)$} $line {} Param value] {
	    error "Syntax error: $line"
	}
	regexp {^"(.*)"$} $value {} value
	set keytype sz
	catch { set keytype [registry type $Key $Param] }
	catch { registry set $Key $Param $value $keytype }
	set status [catch {registry get $Key $Param} tvalue]
	if {$status != 0 || $tvalue != $value} {
	    error "Unable to Set Registry value ($Key $Param=$value)"
	}
    }
}

proc DumpReg {dmpkey} {
    variable REGROOT

    set result {}
    set dkey $REGROOT
    if {[string length $dmpkey] > 0} {
	append dkey \\$dmpkey
    }

    set status [catch {registry keys $dkey} keys ]
    if {$status != 0} {
	return
    }
    
    catch {registry values $dkey} values

    if {[string length $dmpkey]} {
	regsub -all {\\} $dmpkey {.} key
	append result ".$key:\n"
    }

    foreach v $values {
	append result " $v = [registry get $dkey $v]\n"
    }

    foreach k $keys {
	if {[string length $dmpkey]} {
	    append result [DumpReg $dmpkey\\$k]
	} else {
	    append result [DumpReg $k]
	}
    }

    return $result
}

proc DumpRegToFile {Node file} {
    variable REGROOT

    D 2 "CatInfo::DumpRegFromFile $Node $file"

    regexp {^\\(.*)$} $Node {} Node

    set output [DumpReg $Node];

    set tmp $file.[pid]
    file delete -force -- $tmp
    set f [open $tmp w]
    puts -nonewline $f $output
    close $f

    file delete -force -- $file
    file rename -force -- $tmp $file
    D 1 "CatInfo::DumpRegFromFile $Node to $file Done"
    return 1
}
 

# Set path value [delete]
# -- set (or delete) a value, bu full path (starts with a .)
#
proc Set {fullpath value {delete 0}} {
    variable info
    variable REGROOT
    set s {}
    if $delete {
	set s "(delete) "
    }
    D 3 "CatInfo::Set $s$fullpath $value"

    if ![string match .* $fullpath] {
	error "$fullpath: not an absolute path"
    }
    set fullpath [string tolower $fullpath]

 
    regexp {^.*\.([^\.]+)$} $fullpath {} Param
    regexp {^(.*)\.[^\.]+$} $fullpath {} key

    regsub -all {\.} $key {\\} key

    if $delete {
	catch { unset info(data,$fullpath) }
	catch { registry delete $REGROOT$key $Param }
	catch { registry delete $REGROOT$key\\$Param }
    } else {
	registry set $REGROOT\\$key $Param $value
	set info(data,$fullpath) $value
    }
 
    set info(dirty) 1
}

# Get name result
# -- get the value of a value (the name should be absolute)
#    The result is stored in the supplied variable
#    Returns 1 is successful, 0 if not
#
proc Get {name resultvar} {
    upvar $resultvar s
    variable info

    set s {}
    set name [string tolower $name]
    if [info exists info(data,$name)] {
	set s $info(data,$name)
	D 3 "CatInfo::Get $name -> $s"
	return 1
    }
    D 3 "CatInfo::Get $name (not found)"

    return 0
}

# PathGet path name result
# -- Like Get, but searches along a path (a list of nodes).
#    The name should be relative.  If a path variable is given,
#    the path where the variable was found is returned in it.
#
proc PathGet {path name resultvar {pathvar {}}} {
    upvar $resultvar s
    variable info
    if [string length $pathvar] {
	upvar $pathvar p
    }
    D 3 "CatInfo::PathGet $name"
    set s {}
    foreach p $path {
	set p [string tolower $p]
	if [info exists info(data,$p.$name)] {
	    set s $info(data,$p.$name)
	    return 1
	}
    }
    return 0
}

# List
# -- Get all matching paths under a node
#
proc List {Node} {
    variable info

    D 2 "CatInfo::List $Node"
    set node [string tolower $Node]
    set list {}
    foreach name [array names info data,$node.*] {
	regexp {data,(.*)} $name {} path
	lappend list $path
    }
    return [lsort -ascii $list]
}

# Flush
# -- Write any changes back to disk
#
proc Flush {} {
    variable info

    D 1 "CatInfo::Flush"
    if !$info(dirty) {
	return 0
    }
    set info(dirty) 0

    return 1
}

# internal: dump a node back to hte file it was read from
#
proc dump {Node file} {
    variable info

    D 4 "CatInfo::dump $Node $file"
    if ![file exists $file] {
	return 0
    }
    set f [open $file]
    set data [read $f]
    close $f

    set dirty 0
    set output {}
    set node [string tolower $Node]
    set section .
    set Section .
    foreach line [split $data "\n"] {
	set input $line
	set line [string trim $line]
	if ![string length $line] {
	    lappend output $input
	    continue
	}
	if [regexp {^([^#:]+):$} $line {} name] {
	    set Section .$name.
	    set section [string tolower $Section]
	    lappend output $input
	    continue
	}
	if ![regexp {^\s*(#*)\s*(\S+)\s*=\s*(.*)\s*$} $line {} d Param value] {
	    lappend output $input
	    continue
	}
	set param [string tolower $Param]
	if ![info exists info(del,$node$section$param)] {
	    lappend output $input
	    continue
	}
	set d [string match #* $d]
	if $info(del,$node$section$param) {
	    if !$d {
		incr dirty
		set input "# [string trimleft $input]"
	    }
	    lappend output $input
	    continue
	}
	if {$d || [string compare $info(data,$node$section$param) $value]} {
	    incr dirty
	    regsub {^\s*#*\s*([^=]+=\s*)(.*)$} $input \
		    "\\1$info(data,$node$section$param)" input
	    set input "  $input"
	}
	lappend output $input
    }

    if $dirty {
	set tmp $file.[pid]
	file delete -force -- $tmp
	set f [open $tmp w]
	puts -nonewline $f [join $output "\n"]
	close $f

	file delete -force -- $file
	file rename -force -- $tmp $file
	D 4 "CatInfo::dump done"
	return 1
    } else {
	D 4 "CatInfo::dump not dirty"
    }
    return 0
}

# Erase a node, in the form of a disk file,and update the in-memory
# data.
#
proc DeleteNode {Node} {
    variable info

    set node [string tolower $Node]

    if [info exists info(file,$node)] {
	file delete -force $info(file,$node)
	unset info(file,$node)
    }
    truncate $node
    Set $node {} 1
    return 1
}

# Delete everything under a node (in memory)
#
proc truncate {Node} {
    variable info

    D 3 "CatInfo::truncate $Node"
    set node [string tolower $Node]
    foreach prefix {data del name} {
	foreach name [array names info $prefix,$node.*] {
	    unset info($name)
	}
    }
}

# Convert a filename to a node
#
proc filetonode {file} {
    if ![regexp {.*/([^/]+)\.cfg} /$file {} name] {
	return {}
    }
    return [InfoName $name]
}

proc InfoName {name} {
    regsub -all {\.} $name _ name

    return $name
}

}; # CatInfo
