#
# $Id: info.tcl,v 1.18 2000/01/27 16:22:46 elias Exp $
#
# A limited implementation of the catinfo interface.
#
# - Does not handle quoted values.
# - Can only change existing values
# - Can change existing commented-out values, as long as
#   the comment is "simple enough" (a leading # is fine)
# - Will "delete" values by commenting them out
# - Only operates on files 
# - Does not do automagic loading
# - Does very little error-checking.  Almost anything is
#   accepted as a "pathname".  And it is possible to set
#   nodes to values...
#
 

namespace eval ::CatInfo {; #

catch {namespace import CatDbg::*}

namespace export Init Load Flush Set Get PathGet
namespace export Reload LoadFrom List InfoName DeleteNode
 

# The global data array. Contains the following kind of elements:
#
#  data,<path> = value
#  del,<path>  = 0/1	exists => modified, 0/1 => deleted / set
#  file,<path> = file	file associated with a node
#  name,<path> = <Path> <path> is lower-case, <Path> is from the file
#
variable info

variable ROOT
 

# Init
# -- Always call this one first
#
proc Init {rootdir} {
    variable info
    variable ROOT $rootdir
    

    D 1 "CatInfo::Init $rootdir"
    catch {unset info}
    set info(dirty) 0

    if ![file exists $rootdir/global.cfg] {
	return 0
    }
    LoadFrom .global $rootdir/global.cfg
    foreach file [glob -nocomplain $rootdir/services/*.cfg] {
	LoadFrom .services.[filetonode $file] $file
    }
    foreach file [glob -nocomplain $rootdir/tcp/*.cfg] {
	LoadFrom .tcp.[filetonode $file] $file
    }
 
 
    Reload
}

# LoadFrom
# -- Register loading from some place with Reload
#
proc LoadFrom {Node file} {
    variable info

    set node [string tolower $Node]
    D 1 "CatInfo::LoadFrom $Node ($node) $file"
    set info(file,$node) $file
}

# Reload
# -- Re-read all config files that have been loaded so far
#
proc Reload {} {
    variable ROOT
    variable info

    D 1 "CatInfo::Reload"
    foreach x [array names info file,*] {
	regexp {file,(.*)} $x {} path
	truncate $path
	Load $path $info($x)
    }
}

# Load node file
# -- load a file, and hang its data off a certain node
#
proc Load {Node file} {
    variable info

    set f [open $file]
    set data [read $f]
    close $f

    set node [string tolower $Node]
    D 1 "CatInfo::Load $Node $file"
    set info(file,[filetonode $node]) $file

    set section .
    set Section .
    foreach line [split $data "\n"] {
	set line [string trim $line]
	if {![string length $line] || [string match #* $line]} continue
	if [regexp {^([^:]+):$} $line {} name] {
	    set Section .$name.
	    set section [string tolower $Section]
	    continue
	}
	if ![regexp {^(\S+)\s*=\s*(.*)$} $line {} Param value] {
	    error "Zyntaks error: $line"
	}
	set param [string tolower $Param]
	set info(data,$node$section$param) $value
	set info(name,$node$section$param) $Node$Section$Param
	catch { unset info(del,$node$section$param) }
    }
}

 

# Set path value [delete]
# -- set (or delete) a value, bu full path (starts with a .)
#
proc Set {fullpath value {delete 0}} {
    variable info
    set s {}
    if $delete {
	set s "(delete) "
    }
    D 3 "CatInfo::Set $s$fullpath $value"

    if ![string match .* $fullpath] {
	error "$fullpath: not an absolute path"
    }
    set fullpath [string tolower $fullpath]

    if $delete {
	catch { unset info(data,$fullpath) }
    } else {
	set info(data,$fullpath) $value
    }
    set info(del,$fullpath) $delete
 
 
    set info(dirty) 1
}

# Get name result
# -- get the value of a value (the name should be absolute)
#    The result is stored in the supplied variable
#    Returns 1 is successful, 0 if not
#
proc Get {name resultvar} {
    upvar $resultvar s
    variable info

    set s {}
    set name [string tolower $name]
    if [info exists info(data,$name)] {
	set s $info(data,$name)
	D 3 "CatInfo::Get $name -> $s"
	return 1
    }
    D 3 "CatInfo::Get $name (not found)"

    return 0
}

# PathGet path name result
# -- Like Get, but searches along a path (a list of nodes).
#    The name should be relative.  If a path variable is given,
#    the path where the variable was found is returned in it.
#
proc PathGet {path name resultvar {pathvar {}}} {
    upvar $resultvar s
    variable info
    if [string length $pathvar] {
	upvar $pathvar p
    }
    D 3 "CatInfo::PathGet $name"
    set s {}
    foreach p $path {
	set p [string tolower $p]
	if [info exists info(data,$p.$name)] {
	    set s $info(data,$p.$name)
	    return 1
	}
    }
    return 0
}

# List
# -- Get all matching paths under a node
#
proc List {Node} {
    variable info

    D 2 "CatInfo::List $Node"
    set node [string tolower $Node]
    set list {}
    foreach name [array names info data,$node.*] {
	regexp {data,(.*)} $name {} path
	lappend list $path
    }
    return [lsort -ascii $list]
}

# Flush
# -- Write any changes back to disk
#
proc Flush {} {
    variable info

    D 1 "CatInfo::Flush"
    if !$info(dirty) {
	return 0
    }
    foreach name [array names info file,*] {
	set node [lindex [split $name ,] 1]
	dump $node $info($name)
    }
    set info(dirty) 0

    return 1
}

# internal: dump a node back to hte file it was read from
#
proc dump {Node file} {
    variable info

    D 4 "CatInfo::dump $Node $file"
    if ![file exists $file] {
	return 0
    }
    set f [open $file]
    set data [read $f]
    close $f

    set dirty 0
    set output {}
    set node [string tolower $Node]
    set section .
    set Section .
    foreach line [split $data "\n"] {
	set input $line
	set line [string trim $line]
	if ![string length $line] {
	    lappend output $input
	    continue
	}
	if [regexp {^([^#:]+):$} $line {} name] {
	    set Section .$name.
	    set section [string tolower $Section]
	    lappend output $input
	    continue
	}
	if ![regexp {^\s*(#*)\s*(\S+)\s*=\s*(.*)\s*$} $line {} d Param value] {
	    lappend output $input
	    continue
	}
	set param [string tolower $Param]
	if ![info exists info(del,$node$section$param)] {
	    lappend output $input
	    continue
	}
	set d [string match #* $d]
	if $info(del,$node$section$param) {
	    if !$d {
		incr dirty
		set input "# [string trimleft $input]"
	    }
	    lappend output $input
	    continue
	}
	if {$d || [string compare $info(data,$node$section$param) $value]} {
	    incr dirty
	    regsub {^\s*#*\s*([^=]+=\s*)(.*)$} $input \
		    "\\1$info(data,$node$section$param)" input
	    set input "  $input"
	}
	lappend output $input
    }

    if $dirty {
	set tmp $file.[pid]
	file delete -force -- $tmp
	set f [open $tmp w]
	puts -nonewline $f [join $output "\n"]
	close $f

	file delete -force -- $file
	file rename -force -- $tmp $file
	D 4 "CatInfo::dump done"
	return 1
    } else {
	D 4 "CatInfo::dump not dirty"
    }
    return 0
}

# Erase a node, in the form of a disk file,and update the in-memory
# data.
#
proc DeleteNode {Node} {
    variable info

    set node [string tolower $Node]

    if [info exists info(file,$node)] {
	file delete -force $info(file,$node)
	unset info(file,$node)
    }
    truncate $node
    Set $node {} 1
    return 1
}

# Delete everything under a node (in memory)
#
proc truncate {Node} {
    variable info

    D 3 "CatInfo::truncate $Node"
    set node [string tolower $Node]
    foreach prefix {data del name} {
	foreach name [array names info $prefix,$node.*] {
	    unset info($name)
	}
    }
}

# Convert a filename to a node
#
proc filetonode {file} {
    if ![regexp {.*/([^/]+)\.cfg} /$file {} name] {
	return {}
    }
    return [InfoName $name]
}

proc InfoName {name} {
    regsub -all {\.} $name _ name

    return $name
}

}; # CatInfo
