#
# $Id: menutils.tcl,v 1.9 1999/12/27 12:03:51 perre Exp $
#

namespace eval ::Menu { ;#

namespace export PromptList PromptFor GetAnswer GetYesNo DashLine Msg
namespace export Menu pdata

variable pdata

# Check if answer was y/n.
#
proc checkyesno {answer yesvar} {
    upvar $yesvar y
    switch -regexp -- [string tolower $answer] {
	y|ye|yes { set y 1 }
	n|no     { set y 0 }
	default  {
	    puts stderr "Please answer \"y\" or \"n\""
	    return 0
	}
    }
    return 1
}

proc Msg {string {nonewline 0}} {
    puts -nonewline stderr $string
    if !$nonewline {
	puts stderr ""
    }
}

# Generate a line of dashes (or some other character)
#
proc DashLine {len {char -}} {
    set s {}
    for {set i 0} {$i < $len} {incr i} {
	append s $char
    }
    return $s
}

# Prompt for an answer
#
proc GetAnswer {prompt {default {}}} {
    puts -nonewline stderr "$prompt "
    if [string length $default] {
	puts -nonewline stderr "\[$default\] "
    }
    gets stdin s
    set s [string trim $s]
    if ![string length $s] {
	return $default
    }
    return $s
}

# Get a y/n answer. Return boolean
#
proc GetYesNo {prompt {default {}}} {
    while 1 {
	set a [GetAnswer $prompt $default]
	if [checkyesno $a y] {
	    return $y
	}
    }
}


# Present a menu, and run callbacks.
#
# -menu <definition-list>
# -repeat 0/1
# -after <proc>
# -title <string>
#
# The definition list should consist of <char> <proc> <string> ...
# where char triggers the choice, proc is the callback and string the
# description.
#
# The "after" proc is run after each cycle.
#
proc Menu {args} {
    set repeat 0
    set title "Key  Command"
    foreach {flag arg} $args {
	switch -glob -- $flag {
	    -repeat  { set repeat $arg}
	    -after   { set after $arg }
	    -menu    { set list $arg }
	    -title   { set title $arg }
	    default  { error "Unknown flag: $flag" }
	}
    }
    if ![info exists list] {
	return 0
    }
    while 1 {
	puts "\n  $title"
	puts [DashLine 70]
	foreach {c p str} $list {
	    set cmds([string tolower $c]) $p
	    set cmds([string toupper $c]) $p
	    puts "   $c) $str"
	}
	puts [DashLine 70]
	set cmd [GetAnswer "Command:"]
	set cmd [string range $cmd 0 0]
	if [info exists cmds($cmd)] {
	    $cmds($cmd)
	    if [info exists after] {
		$after
	    }
	    if !$repeat break
	} else {
	    Msg "Unknown command: \"$cmd\""
	}
    }
}


# Start a new "prompt list".  The name is used to look up
# text strings in the info array.
#
# If the menu is not aborted, the "then" block is eval:ed.
#
proc PromptList {name info_array resultarray body {then {}}} {
    variable pdata
    upvar $resultarray result

    catch {unset pdata}
    set pdata(name) $name
    set pdata(tags) {}
    set pdata(abort) 0
    set pdata(info_array) $info_array
    uplevel $body
    RunPrompts
    foreach n [array names pdata] {
	set result($n) $pdata($n)
    }
    foreach t $pdata(tags) {
	if ![info exists result($t,value)] {
	    set result($t,value) {}
	}
    }
    if {!$pdata(abort) && [string length $then]} {
	uplevel $then
    }
}

# Add a prompt to the current prompt list.
#
#  tag:   identifies the entry.
#  type:  y - yes/no
#         s - string
#         z - non-empty string
#         n - numeric
#         m - multiple-choice, requires -data arg
#
#  prompt: prompt...
#  args:  -default value
#         -help    help-string
#	  -value   current-value
#         -data    multiple-choice-list { value string-to-show }
#         -proc    validation-proc {arrayvar tag value}
#	  -enable  tag-list (only for type y). Turn on/off other entries
#	  -disable tag-list
#         -off     turn off the current prompt.
#
proc PromptFor {tag type prompt args} {
    variable pdata

    lappend pdata(tags) $tag
    set pdata($tag,type) $type
    set pdata($tag,prompt) $prompt
    set pdata($tag,enabled) 1
    set pdata($tag,default) {}
    set pdata($tag,value) {}
    foreach {flag arg} $args {
	switch -glob -- $flag {
	    -def* { set pdata($tag,default) $arg }
	    -hel* { set pdata($tag,help) $arg }
	    -dat* { set pdata($tag,data) $arg }
	    -pro* { set pdata($tag,proc) $arg }
	    -val* { set pdata($tag,oldvalue) $arg }
	    -ena* { set pdata($tag,enables) $arg }
	    -dis* { set pdata($tag,disables) $arg }
	    -off* { set pdata($tag,enabled) $arg }
	    default { error "Unknown flag: $flag" }
	}
    }
}

# Run the currently configured prompts, and return the results in
# the supplied array variable. Elements look like: <tag>,<param>
# for parameters: type, prompt, enabled, default, help, data,
# proc, disables, value 
#
proc RunPrompts {} {
    variable pdata
    upvar 2 $pdata(info_array) di

    foreach tag $pdata(tags) {
	if $pdata(abort) return
	if !$pdata($tag,enabled) continue
	if {"$pdata($tag,type)" == "m"} {
	    while 1 {
		MultipleChoice $tag di
		if ![HandleAnswer $tag $pdata($tag,value)] continue
		if $pdata(abort) return
		break
	    }
	    continue
	}
	while 1 {
	    set a [GetAnswer $pdata($tag,prompt) $pdata($tag,default)]
	    if {[string compare ? $a] == 0} {
		set help {}
		if [info exists pdata($tag,help)] {
		    set help "$pdata($tag,help)\n"
		} elseif [info exists di($pdata(name),$tag,help)] {
		    set help "$di($pdata(name),$tag,help)\n"
		}
		if [info exists pdata($tag,oldvalue)] {
		    append help "\nCurrent value: \"$pdata($tag,oldvalue)\""
		}
		if ![string length $help] {
		    puts stderr "No help available"
		    continue
		}
		puts stderr $help
	    } else {
		if ![PromptTypeOk $tag $a] continue
		if ![HandleAnswer $tag $pdata($tag,value)] continue
		break
	    }
	    if $pdata(abort) return
	}
    }
}

# Check if a response made sense, according to the type
#
proc PromptTypeOk {tag answer} {
    variable pdata

    switch $pdata($tag,type) {
	s {
	}
	z {
	    if ![string length $answer] {
		puts stderr "Answer cannot be empty"
		return 0
	    }
	}
	n {
	    if {[scan $answer %d n] != 1 || $n < 0} {
		puts stderr "Answer should be a positive number"
		return 0
	    }
	}
	y {
	    if [checkyesno $answer y] {
		set pdata($tag,value) $y
		return 1
	    }
	    return 0
	}
    }
    set pdata($tag,value) $answer

    return 1
}

# Run a multiple-choice question
#
proc MultipleChoice {tag info_array} {
    variable pdata
    upvar $info_array di

    while 1 {
	puts stderr "\n$pdata($tag,prompt)\n[DashLine 70 -]"
	set i 0
	foreach {value line} $pdata($tag,data) {
	    incr i
	    set values($i) $value
	    puts stderr [format " %2d. %s" $i $line]
	}
	puts stderr [DashLine 70 -]
	if {$i == 1} {
	    set a [GetAnswer "Selection: " $i]
	} else {
	    set a [GetAnswer "Selection (1 - $i):"]
	}
	if {[scan $a %u c] == 1 && $c > 0 && $c <= $i} {
	    set pdata($tag,value) $values($c)
	    return 1
	} elseif {[string compare $a ?] == 0} {
	    set help {}
	    if [info exists pdata($tag,help)] {
		set help "$pdata($tag,help)\n"
	    } elseif [info exists di($pdata(name),$tag,help)] {
		set help "$di($pdata(name),$tag,help)\n"
	    }
#	    if [info exists pdata($tag,oldvalue)] {
#		append help "\nCurrent value: \"$pdata($tag,oldvalue)\""
#	    }
	    if ![string length $help] {
		puts stderr "No help available"
		continue
	    }
	    puts stderr $help
	} else {
	    puts stderr "\nBad choice: \"$a\"\n"
	}
    }
}

# React on input (run callbacks, enable/disable other menus etc)
#
proc HandleAnswer {tag answer} {
    variable pdata

    if [info exists pdata($tag,enables)] {
	foreach t $pdata($tag,enables) {
	    set pdata($t,enabled) $answer
	}
    }
    if [info exists pdata($tag,disables)] {
	foreach t $pdata($tag,disables) {	
	    set pdata($t,enabled) [expr !$answer]
	}
    }
    if [info exists pdata($tag,proc)] {
	return [$pdata($tag,proc) pdata $tag $answer]
    }
    return 1
}

} ;# end namespace Menu

namespace eval ::ParamConfig { ;#

# Take care of agent specific configuration. This is an interface to adding
# agent specific confuguration parameters to the info tree.
#
proc parameter { tag type info_path prompt } {
    variable ::Menu::pdata

    set gotit 0

    for { set no 0 } { $no < 20 } { incr no } {
       if [info exist pdata(custom_$no,tag)] {
	   if ![string compare $pdata(custom_$no,tag) $tag] {
	       set pdata(custom_$no,type) $type
	       set pdata(custom_$no,enabled) 1
	       set pdata(custom_$no,prompt) $prompt
	       set pdata(custom_$no,path) $info_path	
 	       set gotit 1
           }
       }
    }

    if { !$gotit } {
	if [info exist pdata(custom_value)] {
	    incr pdata(custom_value)
	} else {
	    set pdata(custom_value) 0
	}
        set no $pdata(custom_value)
	set pdata(custom_$no,tag) $tag
	set pdata(custom_$no,type) $type
	set pdata(custom_$no,enabled) 1
	set pdata(custom_$no,prompt) $prompt
	set pdata(custom_$no,path) $info_path	
    } 
}

proc help { tag help_message } {
    variable ::Menu::pdata

    set gotit 0

    for { set no 0 } { $no < 20 } {incr no } { 
       if [info exist pdata(custom_$no,tag)] {
  	  if ![string compare $pdata(custom_$no,tag) $tag] {
	     set pdata(custom_$no,help) $help_message
             set gotit 1
	  }
       }
    }

    if { !$gotit } {
	if [info exist pdata(custom_value)] {
	    incr pdata(custom_value)
	} else {
	    set pdata(custom_value) 0
	}
        set no $pdata(custom_value)
	set pdata(custom_$no,tag) $tag
	set pdata(custom_$no,help) $help_message
    } 
}

} ;# end namespace ParamConfig


